@@ -36,6 +36,7 @@ Created 1/8/1996 Heikki Tuuri
 #include "row0mysql.h"
 #include "pars0pars.h"
 #include "trx0roll.h"
+#include "trx0rseg.h"
 #include "trx0undo.h"
 #include "ut0vec.h"
 #include "dict0priv.h"
@@ -62,7 +63,9 @@ dict_create_sys_tables_tuple(
 	ulint		type;
 
 	ut_ad(table);
+	ut_ad(!table->space || table->space->id == table->space_id);
 	ut_ad(heap);
+	ut_ad(table->n_cols >= DATA_N_SYS_COLS);
 
 	sys_tables = dict_sys->sys_tables;
 
@@ -96,12 +99,11 @@ dict_create_sys_tables_tuple(
 
 	/* If there is any virtual column, encode it in N_COLS */
 	mach_write_to_4(ptr, dict_table_encode_n_col(
-				static_cast<ulint>(table->n_def),
-				static_cast<ulint>(table->n_v_def))
-			| ((table->flags & DICT_TF_COMPACT) << 31));
+				ulint(table->n_cols - DATA_N_SYS_COLS),
+				ulint(table->n_v_def))
+			| (ulint(table->flags & DICT_TF_COMPACT) << 31));
 	dfield_set_data(dfield, ptr, 4);
 
-
 	/* 5: TYPE (table flags) -----------------------------*/
 	dfield = dtuple_get_nth_field(
 		entry, DICT_COL__SYS_TABLES__TYPE);
@@ -145,7 +147,7 @@ dict_create_sys_tables_tuple(
 		entry, DICT_COL__SYS_TABLES__SPACE);
 
 	ptr = static_cast<byte*>(mem_heap_alloc(heap, 4));
-	mach_write_to_4(ptr, table->space);
+	mach_write_to_4(ptr, table->space_id);
 
 	dfield_set_data(dfield, ptr, 4);
 	/*----------------------------------*/
@@ -188,7 +190,7 @@ dict_create_sys_columns_tuple(
 		v_col_no = column->ind;
 	} else {
 		column = dict_table_get_nth_col(table, i);
-		ut_ad(!dict_col_is_virtual(column));
+		ut_ad(!column->is_virtual());
 	}
 
 	sys_columns = dict_sys->sys_columns;
@@ -348,15 +350,14 @@ dict_build_table_def_step(
 	que_thr_t*	thr,	/*!< in: query thread */
 	tab_node_t*	node)	/*!< in: table create node */
 {
-	dict_table_t*	table;
-
-	table = node->table;
-	ut_ad(!dict_table_is_temporary(table));
-
-	trx_t*	trx = thr_get_trx(thr);
-	dict_table_assign_new_id(table, trx);
-
 	ut_ad(mutex_own(&dict_sys->mutex));
+	dict_table_t*	table = node->table;
+	trx_t* trx = thr_get_trx(thr);
+	ut_ad(!table->is_temporary());
+	ut_ad(!table->space);
+	ut_ad(table->space_id == ULINT_UNDEFINED);
+	dict_hdr_get_new_id(&table->id, NULL, NULL);
+	trx->table_id = table->id;
 
 	/* Always set this bit for all new created tables */
 	DICT_TF2_FLAG_SET(table, DICT_TF2_FTS_AUX_HEX_NAME);
@@ -367,12 +368,10 @@ dict_build_table_def_step(
 	if (DICT_TF2_FLAG_IS_SET(table, DICT_TF2_USE_FILE_PER_TABLE)) {
 		/* This table will need a new tablespace. */
 
-		ut_ad(dict_table_get_format(table) <= UNIV_FORMAT_MAX);
 		ut_ad(DICT_TF_GET_ZIP_SSIZE(table->flags) == 0
-		      || dict_table_get_format(table) >= UNIV_FORMAT_B);
-		ut_ad(trx->table_id);
+		      || dict_table_has_atomic_blobs(table));
 		mtr_t mtr;
-		trx_undo_t* undo = trx->rsegs.m_redo.insert_undo;
+		trx_undo_t* undo = trx->rsegs.m_redo.undo;
 		if (undo && !undo->table_id
 		    && trx_get_dict_operation(trx) == TRX_DICT_OP_TABLE) {
 			/* This must be a TRUNCATE operation where
@@ -381,41 +380,43 @@ dict_build_table_def_step(
 			associated with the new empty table, so that
 			we can remove it on recovery. */
 			mtr.start();
-			trx_undo_mark_as_dict(trx, undo, &mtr);
+			undo->table_id = trx->table_id;
+			undo->dict_operation = TRUE;
+			page_t* page = trx_undo_page_get(
+				page_id_t(trx->rsegs.m_redo.rseg->space->id,
+					  undo->hdr_page_no),
+				&mtr);
+			mlog_write_ulint(page + undo->hdr_offset
+					 + TRX_UNDO_DICT_TRANS,
+					 TRUE, MLOG_1BYTE, &mtr);
+			mlog_write_ull(page + undo->hdr_offset
+				       + TRX_UNDO_TABLE_ID,
+				       trx->table_id, &mtr);
 			mtr.commit();
 			log_write_up_to(mtr.commit_lsn(), true);
 		}
-		ulint	space;
 		/* Get a new tablespace ID */
-		dict_hdr_get_new_id(NULL, NULL, &space, table, false);
+		ulint space_id;
+		dict_hdr_get_new_id(NULL, NULL, &space_id);
 
 		DBUG_EXECUTE_IF(
 			"ib_create_table_fail_out_of_space_ids",
-			space = ULINT_UNDEFINED;
+			space_id = ULINT_UNDEFINED;
 		);
 
-		if (space == ULINT_UNDEFINED) {
+		if (space_id == ULINT_UNDEFINED) {
 			return DB_ERROR;
 		}
-		table->space = unsigned(space);
 
 		/* Determine the tablespace flags. */
 		bool	has_data_dir = DICT_TF_HAS_DATA_DIR(table->flags);
 		ulint	fsp_flags = dict_tf_to_fsp_flags(table->flags);
-		char*	filepath;
-
-		if (has_data_dir) {
-			ut_ad(table->data_dir_path);
-			filepath = fil_make_filepath(
-				table->data_dir_path,
-				table->name.m_name, IBD, true);
-
-		} else {
-			/* Make the tablespace file in the default dir
-			using the table name */
-			filepath = fil_make_filepath(
-				NULL, table->name.m_name, IBD, false);
-		}
+		ut_ad(!has_data_dir || table->data_dir_path);
+		char*	filepath = has_data_dir
+			? fil_make_filepath(table->data_dir_path,
+					    table->name.m_name, IBD, true)
+			: fil_make_filepath(NULL,
+					    table->name.m_name, IBD, false);
 
 		/* We create a new single-table tablespace for the table.
 		We initially let it be 4 pages:
@@ -425,49 +426,36 @@ dict_build_table_def_step(
 		- page 3 will contain the root of the clustered index of
 		the table we create here. */
 
-		dberr_t err = fil_ibd_create(
-			space, table->name.m_name, filepath, fsp_flags,
-			FIL_IBD_FILE_INITIAL_SIZE, node->mode, node->key_id);
+		dberr_t err;
+		table->space = fil_ibd_create(
+			space_id, table->name.m_name, filepath, fsp_flags,
+			FIL_IBD_FILE_INITIAL_SIZE,
+			node->mode, node->key_id, &err);
 
 		ut_free(filepath);
 
-		if (err != DB_SUCCESS) {
+		if (!table->space) {
+			ut_ad(err != DB_SUCCESS);
 			return err;
 		}
 
+		table->space_id = space_id;
 		mtr.start();
 		mtr.set_named_space(table->space);
-
 		fsp_header_init(table->space, FIL_IBD_FILE_INITIAL_SIZE, &mtr);
-
 		mtr.commit();
 	} else {
 		ut_ad(dict_tf_get_rec_format(table->flags)
 		      != REC_FORMAT_COMPRESSED);
-		ut_ad(table->space == srv_sys_space.space_id());
+		table->space = fil_system.sys_space;
+		table->space_id = TRX_SYS_SPACE;
 	}
 
 	ins_node_set_new_row(node->tab_def,
 			     dict_create_sys_tables_tuple(table, node->heap));
-
 	return DB_SUCCESS;
 }
 
-/***************************************************************//**
-Builds a column definition to insert. */
-static
-void
-dict_build_col_def_step(
-/*====================*/
-	tab_node_t*	node)	/*!< in: table create node */
-{
-	dtuple_t*	row;
-
-	row = dict_create_sys_columns_tuple(node->table, node->col_no,
-					    node->heap);
-	ins_node_set_new_row(node->col_def, row);
-}
-
 /** Builds a SYS_VIRTUAL row definition to insert.
 @param[in]	node	table create node */
 static
@@ -497,19 +485,19 @@ dict_create_sys_indexes_tuple(
 					tuple is allocated */
 {
 	dict_table_t*	sys_indexes;
-	dict_table_t*	table;
 	dtuple_t*	entry;
 	dfield_t*	dfield;
 	byte*		ptr;
 
 	ut_ad(mutex_own(&dict_sys->mutex));
 	ut_ad(index);
+	ut_ad(index->table->space || index->table->file_unreadable);
+	ut_ad(!index->table->space
+	      || index->table->space->id == index->table->space_id);
 	ut_ad(heap);
 
 	sys_indexes = dict_sys->sys_indexes;
 
-	table = dict_table_get_low(index->table_name);
-
 	entry = dtuple_create(
 		heap, DICT_NUM_COLS__SYS_INDEXES + DATA_N_SYS_COLS);
 
@@ -520,7 +508,7 @@ dict_create_sys_indexes_tuple(
 		entry, DICT_COL__SYS_INDEXES__TABLE_ID);
 
 	ptr = static_cast<byte*>(mem_heap_alloc(heap, 8));
-	mach_write_to_8(ptr, table->id);
+	mach_write_to_8(ptr, index->table->id);
 
 	dfield_set_data(dfield, ptr, 8);
 
@@ -574,7 +562,7 @@ dict_create_sys_indexes_tuple(
 		entry, DICT_COL__SYS_INDEXES__SPACE);
 
 	ptr = static_cast<byte*>(mem_heap_alloc(heap, 4));
-	mach_write_to_4(ptr, index->space);
+	mach_write_to_4(ptr, index->table->space_id);
 
 	dfield_set_data(dfield, ptr, 4);
 
@@ -742,7 +730,8 @@ dict_build_index_def_step(
 
 	index = node->index;
 
-	table = dict_table_get_low(index->table_name);
+	table = index->table = node->table = dict_table_open_on_name(
+		node->table_name, TRUE, FALSE, DICT_ERR_IGNORE_NONE);
 
 	if (table == NULL) {
 		return(DB_TABLE_NOT_FOUND);
@@ -753,17 +742,14 @@ dict_build_index_def_step(
 		trx->table_id = table->id;
 	}
 
-	node->table = table;
-
 	ut_ad((UT_LIST_GET_LEN(table->indexes) > 0)
 	      || dict_index_is_clust(index));
 
-	dict_hdr_get_new_id(NULL, &index->id, NULL, table, false);
+	dict_hdr_get_new_id(NULL, &index->id, NULL);
 
 	/* Inherit the space id from the table; we store all indexes of a
 	table in the same tablespace */
 
-	index->space = table->space;
 	node->page_no = FIL_NULL;
 	row = dict_create_sys_indexes_tuple(index, node->heap);
 	node->ind_row = row;
@@ -774,6 +760,7 @@ dict_build_index_def_step(
 	index->trx_id = trx->id;
 	ut_ad(table->def_trx_id <= trx->id);
 	table->def_trx_id = trx->id;
+	dict_table_close(table, true, false);
 
 	return(DB_SUCCESS);
 }
@@ -798,12 +785,7 @@ dict_build_index_def(
 	ut_ad((UT_LIST_GET_LEN(table->indexes) > 0)
 	      || dict_index_is_clust(index));
 
-	dict_hdr_get_new_id(NULL, &index->id, NULL, table, false);
-
-	/* Inherit the space id from the table; we store all indexes of a
-	table in the same tablespace */
-
-	index->space = table->space;
+	dict_hdr_get_new_id(NULL, &index->id, NULL);
 
 	/* Note that the index was created by this transaction. */
 	index->trx_id = trx->id;
@@ -857,14 +839,7 @@ dict_create_index_tree_step(
 	the index and its root address is written to the index entry in
 	sys_indexes */
 
-	mtr_start(&mtr);
-
-	const bool	missing = !index->is_readable()
-		|| dict_table_is_discarded(index->table);
-
-	if (!missing) {
-		mtr.set_named_space(index->space);
-	}
+	mtr.start();
 
 	search_tuple = dict_create_search_tuple(node->ind_row, node->heap);
 
@@ -877,13 +852,14 @@ dict_create_index_tree_step(
 
 	dberr_t		err = DB_SUCCESS;
 
-	if (missing) {
+	if (!index->is_readable()) {
 		node->page_no = FIL_NULL;
 	} else {
+		index->set_modified(mtr);
+
 		node->page_no = btr_create(
-			index->type, index->space,
-			dict_table_page_size(index->table),
-			index->id, index, NULL, &mtr);
+			index->type, index->table->space,
+			index->id, index, &mtr);
 
 		if (node->page_no == FIL_NULL) {
 			err = DB_OUT_OF_FILE_SPACE;
@@ -900,7 +876,7 @@ dict_create_index_tree_step(
 
 	btr_pcur_close(&pcur);
 
-	mtr_commit(&mtr);
+	mtr.commit();
 
 	return(err);
 }
@@ -916,7 +892,6 @@ dict_create_index_tree_in_mem(
 	const trx_t*	trx)	/*!< in: InnoDB transaction handle */
 {
 	mtr_t		mtr;
-	ulint		page_no;
 
 	ut_ad(mutex_own(&dict_sys->mutex));
 	ut_ad(!(index->type & DICT_FTS));
@@ -924,28 +899,18 @@ dict_create_index_tree_in_mem(
 	mtr_start(&mtr);
 	mtr_set_log_mode(&mtr, MTR_LOG_NO_REDO);
 
-	dberr_t		err = DB_SUCCESS;
-
 	/* Currently this function is being used by temp-tables only.
 	Import/Discard of temp-table is blocked and so this assert. */
 	ut_ad(index->is_readable());
-	ut_ad(!dict_table_is_discarded(index->table));
-
-	page_no = btr_create(
-		index->type, index->space,
-		dict_table_page_size(index->table),
-		index->id, index, NULL, &mtr);
-
-	index->page = page_no;
-	index->trx_id = trx->id;
-
-	if (page_no == FIL_NULL) {
-		err = DB_OUT_OF_FILE_SPACE;
-	}
+	ut_ad(!(index->table->flags2 & DICT_TF2_DISCARDED));
 
+	index->page = btr_create(index->type, index->table->space,
+				 index->id, index, &mtr);
 	mtr_commit(&mtr);
 
-	return(err);
+	index->trx_id = trx->id;
+
+	return index->page == FIL_NULL ? DB_OUT_OF_FILE_SPACE : DB_SUCCESS;
 }
 
 /** Drop the index tree associated with a row in SYS_INDEXES table.
@@ -1006,44 +971,12 @@ dict_drop_index_tree(
 		return(false);
 	}
 
-	/* If tablespace is scheduled for truncate, do not try to drop
-	the indexes in that tablespace. There is a truncate fixup action
-	which will take care of it. */
-	if (srv_is_tablespace_truncated(space)) {
-		return(false);
-	}
-
 	btr_free_if_exists(page_id_t(space, root_page_no), page_size,
 			   mach_read_from_8(ptr), mtr);
 
 	return(true);
 }
 
-/*******************************************************************//**
-Drops the index tree but don't update SYS_INDEXES table. */
-void
-dict_drop_index_tree_in_mem(
-/*========================*/
-	const dict_index_t*	index,		/*!< in: index */
-	ulint			page_no)	/*!< in: index page-no */
-{
-	ut_ad(mutex_own(&dict_sys->mutex));
-	ut_ad(dict_table_is_temporary(index->table));
-
-	ulint			root_page_no = page_no;
-	ulint			space = index->space;
-	bool			found;
-	const page_size_t	page_size(fil_space_get_page_size(space,
-								  &found));
-
-	/* If tree has already been freed or it is a single table
-	tablespace and the .ibd file is missing do nothing,
-	else free the all the pages */
-	if (root_page_no != FIL_NULL && found) {
-		btr_free(page_id_t(space, root_page_no), page_size);
-	}
-}
-
 /*******************************************************************//**
 Recreate the index tree associated with a row in SYS_INDEXES table.
 @return	new root page number, or FIL_NULL on failure */
@@ -1061,28 +994,22 @@ dict_recreate_index_tree(
 {
 	ut_ad(mutex_own(&dict_sys->mutex));
 	ut_a(!dict_table_is_comp(dict_sys->sys_indexes));
+	ut_ad(!table->space || table->space->id == table->space_id);
 
 	ulint		len;
-	rec_t*		rec = btr_pcur_get_rec(pcur);
+	const rec_t*	rec = btr_pcur_get_rec(pcur);
 
 	const byte*	ptr = rec_get_nth_field_old(
 		rec, DICT_FLD__SYS_INDEXES__PAGE_NO, &len);
 
 	ut_ad(len == 4);
 
-	ulint	root_page_no = mtr_read_ulint(ptr, MLOG_4BYTES, mtr);
-
-	ptr = rec_get_nth_field_old(rec, DICT_FLD__SYS_INDEXES__SPACE, &len);
+	ut_ad(table->space_id == mach_read_from_4(
+		      rec_get_nth_field_old(rec, DICT_FLD__SYS_INDEXES__SPACE,
+					    &len)));
 	ut_ad(len == 4);
 
-	ut_a(table->space == mtr_read_ulint(ptr, MLOG_4BYTES, mtr));
-
-	ulint			space = table->space;
-	bool			found;
-	const page_size_t	page_size(fil_space_get_page_size(space,
-								  &found));
-
-	if (!found) {
+	if (!table->space) {
 		/* It is a single table tablespae and the .ibd file is
 		missing: do nothing. */
 
@@ -1108,7 +1035,7 @@ dict_recreate_index_tree(
 	mtr_commit(mtr);
 
 	mtr_start(mtr);
-	mtr->set_named_space(space);
+	mtr->set_named_space(table->space);
 	btr_pcur_restore_position(BTR_MODIFY_LEAF, pcur, mtr);
 
 	/* Find the index corresponding to this SYS_INDEXES record. */
@@ -1116,15 +1043,12 @@ dict_recreate_index_tree(
 	     index != NULL;
 	     index = UT_LIST_GET_NEXT(indexes, index)) {
 		if (index->id == index_id) {
-			if (index->type & DICT_FTS) {
-				return(FIL_NULL);
-			} else {
-				root_page_no = btr_create(
-					type, space, page_size, index_id,
-					index, NULL, mtr);
-				index->page = (unsigned int) root_page_no;
-				return(root_page_no);
-			}
+			ulint root_page_no = (index->type & DICT_FTS)
+				? FIL_NULL
+				: btr_create(type, table->space,
+					     index_id, index, mtr);
+			index->page = unsigned(root_page_no);
+			return root_page_no;
 		}
 	}
 
@@ -1134,73 +1058,6 @@ dict_recreate_index_tree(
 	return(FIL_NULL);
 }
 
-/*******************************************************************//**
-Truncates the index tree but don't update SYSTEM TABLES.
-@return DB_SUCCESS or error */
-dberr_t
-dict_truncate_index_tree_in_mem(
-/*============================*/
-	dict_index_t*	index)		/*!< in/out: index */
-{
-	mtr_t		mtr;
-	bool		truncate;
-	ulint		space = index->space;
-
-	ut_ad(mutex_own(&dict_sys->mutex));
-	ut_ad(dict_table_is_temporary(index->table));
-
-	ulint		type = index->type;
-	ulint		root_page_no = index->page;
-
-	if (root_page_no == FIL_NULL) {
-
-		/* The tree has been freed. */
-		ib::warn() << "Trying to TRUNCATE a missing index of table "
-			<< index->table->name << "!";
-
-		truncate = false;
-	} else {
-		truncate = true;
-	}
-
-	bool			found;
-	const page_size_t	page_size(fil_space_get_page_size(space,
-								  &found));
-
-	if (!found) {
-
-		/* It is a single table tablespace and the .ibd file is
-		missing: do nothing */
-
-		ib::warn()
-			<< "Trying to TRUNCATE a missing .ibd file of table "
-			<< index->table->name << "!";
-	}
-
-	/* If table to truncate resides in its on own tablespace that will
-	be re-created on truncate then we can ignore freeing of existing
-	tablespace objects. */
-
-	if (truncate) {
-		btr_free(page_id_t(space, root_page_no), page_size);
-	}
-
-	mtr_start(&mtr);
-	mtr_set_log_mode(&mtr, MTR_LOG_NO_REDO);
-
-	root_page_no = btr_create(
-		type, space, page_size, index->id, index, NULL, &mtr);
-
-	DBUG_EXECUTE_IF("ib_err_trunc_temp_recreate_index",
-			root_page_no = FIL_NULL;);
-
-	index->page = root_page_no;
-
-	mtr_commit(&mtr);
-
-	return(index->page == FIL_NULL ? DB_ERROR : DB_SUCCESS);
-}
-
 /*********************************************************************//**
 Creates a table create graph.
 @return own: table create node */
@@ -1244,6 +1101,7 @@ tab_create_graph_create(
 
 /** Creates an index create graph.
 @param[in]	index	index to create, built as a memory data structure
+@param[in]	table	table name
 @param[in,out]	heap	heap where created
 @param[in]	add_v	new virtual columns added in the same clause with
 			add index
@@ -1251,6 +1109,7 @@ tab_create_graph_create(
 ind_node_t*
 ind_create_graph_create(
 	dict_index_t*		index,
+	const char*		table,
 	mem_heap_t*		heap,
 	const dict_add_v_col_t*	add_v)
 {
@@ -1263,6 +1122,8 @@ ind_create_graph_create(
 
 	node->index = index;
 
+	node->table_name = table;
+
 	node->add_v = add_v;
 
 	node->state = INDEX_BUILD_INDEX_DEF;
@@ -1325,12 +1186,19 @@ dict_create_table_step(
 
 	if (node->state == TABLE_BUILD_COL_DEF) {
 
-		if (node->col_no < (static_cast<ulint>(node->table->n_def)
-				    + static_cast<ulint>(node->table->n_v_def))) {
+		if (node->col_no + DATA_N_SYS_COLS
+		    < (static_cast<ulint>(node->table->n_def)
+		       + static_cast<ulint>(node->table->n_v_def))) {
 
-			dict_build_col_def_step(node);
+			ulint i = node->col_no++;
+			if (i + DATA_N_SYS_COLS >= node->table->n_def) {
+				i += DATA_N_SYS_COLS;
+			}
 
-			node->col_no++;
+			ins_node_set_new_row(
+				node->col_def,
+				dict_create_sys_columns_tuple(node->table, i,
+							      node->heap));
 
 			thr->run_node = node->col_def;
 
@@ -1388,7 +1256,8 @@ dict_create_table_step(
 	if (node->state == TABLE_ADD_TO_CACHE) {
 		DBUG_EXECUTE_IF("ib_ddl_crash_during_create", DBUG_SUICIDE(););
 
-		dict_table_add_to_cache(node->table, TRUE, node->heap);
+		node->table->can_be_evicted = true;
+		node->table->add_to_cache();
 
 		err = DB_SUCCESS;
 	}
@@ -1473,21 +1342,26 @@ dict_create_index_step(
 	}
 
 	if (node->state == INDEX_ADD_TO_CACHE) {
+		ut_ad(node->index->table == node->table);
+		node->index = dict_index_add_to_cache(
+			node->index, FIL_NULL, trx_is_strict(trx),
+			&err, node->add_v);
 
-		index_id_t	index_id = node->index->id;
-
-		err = dict_index_add_to_cache_w_vcol(
-			node->table, node->index, node->add_v, FIL_NULL,
-			trx_is_strict(trx));
-
-		node->index = dict_index_get_if_in_cache_low(index_id);
-		ut_a((node->index == NULL) == (err != DB_SUCCESS));
-
-		if (err != DB_SUCCESS) {
+		ut_ad((node->index == NULL) == (err != DB_SUCCESS));
 
+		if (!node->index) {
 			goto function_exit;
 		}
 
+		ut_ad(!node->index->is_instant());
+		ut_ad(node->index->n_core_null_bytes
+		      == ((dict_index_is_clust(node->index)
+			   && node->table->supports_instant())
+			  ? dict_index_t::NO_CORE_NULL_BYTES
+			  : UT_BITS_IN_BYTES(
+				  unsigned(node->index->n_nullable))));
+		node->index->n_core_null_bytes = UT_BITS_IN_BYTES(
+			unsigned(node->index->n_nullable));
 		node->state = INDEX_CREATE_INDEX_TREE;
 	}
 
@@ -1639,7 +1513,7 @@ dict_create_or_check_foreign_constraint_tables(void)
 		return(DB_READ_ONLY);
 	}
 
-	trx = trx_allocate_for_mysql();
+	trx = trx_create();
 
 	trx_set_dict_operation(trx, TRX_DICT_OP_TABLE);
 
@@ -1733,7 +1607,7 @@ dict_create_or_check_foreign_constraint_tables(void)
 
 	row_mysql_unlock_data_dictionary(trx);
 
-	trx_free_for_mysql(trx);
+	trx_free(trx);
 
 	srv_file_per_table = srv_file_per_table_backup;
 
@@ -1779,7 +1653,7 @@ dict_create_or_check_sys_virtual()
 		return(DB_READ_ONLY);
 	}
 
-	trx = trx_allocate_for_mysql();
+	trx = trx_create();
 
 	trx_set_dict_operation(trx, TRX_DICT_OP_TABLE);
 
@@ -1835,7 +1709,7 @@ dict_create_or_check_sys_virtual()
 
 	row_mysql_unlock_data_dictionary(trx);
 
-	trx_free_for_mysql(trx);
+	trx_free(trx);
 
 	srv_file_per_table = srv_file_per_table_backup;
 
@@ -2074,7 +1948,8 @@ dict_create_add_foreign_to_dictionary(
 				  foreign->referenced_table_name);
 
 	pars_info_add_int4_literal(info, "n_cols",
-				   foreign->n_fields + (foreign->type << 24));
+				   ulint(foreign->n_fields)
+				   | (ulint(foreign->type) << 24));
 
 	DBUG_PRINT("dict_create_add_foreign_to_dictionary",
 		   ("'%s', '%s', '%s', %d", foreign->id, name,
@@ -2252,6 +2127,8 @@ dict_create_add_foreigns_to_dictionary(
 		return(DB_ERROR);
 	}
 
+	error = DB_SUCCESS;
+
 	for (dict_foreign_set::const_iterator it = local_fk_set.begin();
 	     it != local_fk_set.end();
 	     ++it) {
@@ -2263,12 +2140,11 @@ dict_create_add_foreigns_to_dictionary(
 			table->name.m_name, foreign, trx);
 
 		if (error != DB_SUCCESS) {
-
-			return(error);
+			break;
 		}
 	}
 
-	return(DB_SUCCESS);
+	return error;
 }
 
 /****************************************************************//**
@@ -2306,7 +2182,7 @@ dict_create_or_check_sys_tablespace(void)
 		return(DB_READ_ONLY);
 	}
 
-	trx = trx_allocate_for_mysql();
+	trx = trx_create();
 
 	trx_set_dict_operation(trx, TRX_DICT_OP_TABLE);
 
@@ -2368,7 +2244,7 @@ dict_create_or_check_sys_tablespace(void)
 
 	row_mysql_unlock_data_dictionary(trx);
 
-	trx_free_for_mysql(trx);
+	trx_free(trx);
 
 	srv_file_per_table = srv_file_per_table_backup;
 
@@ -2500,15 +2376,3 @@ dict_delete_tablespace_and_datafiles(
 
 	return(err);
 }
-
-/** Assign a new table ID and put it into the table cache and the transaction.
-@param[in,out]	table	Table that needs an ID
-@param[in,out]	trx	Transaction */
-void
-dict_table_assign_new_id(
-	dict_table_t*	table,
-	trx_t*		trx)
-{
-	dict_hdr_get_new_id(&table->id, NULL, NULL, table, false);
-	trx->table_id = table->id;
-}
