@@ -1,7 +1,7 @@
 /*****************************************************************************
 
 Copyright (c) 1995, 2016, Oracle and/or its affiliates. All Rights Reserved.
-Copyright (c) 2017, 2020, MariaDB Corporation.
+Copyright (c) 2017, 2018, MariaDB Corporation.
 
 This program is free software; you can redistribute it and/or modify it under
 the terms of the GNU General Public License as published by the Free Software
@@ -13,7 +13,7 @@ FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License along with
 this program; if not, write to the Free Software Foundation, Inc.,
-51 Franklin Street, Fifth Floor, Boston, MA 02110-1335 USA
+51 Franklin Street, Suite 500, Boston, MA 02110-1335 USA
 
 *****************************************************************************/
 
@@ -408,7 +408,7 @@ xdes_get_descriptor_with_space_hdr(
 	ulint	size;
 	ulint	descr_page_no;
 	page_t*	descr_page;
-	ut_ad(mtr_memo_contains(mtr, space, MTR_MEMO_SPACE_X_LOCK));
+	ut_ad(mtr_memo_contains(mtr, &space->latch, MTR_MEMO_X_LOCK));
 	ut_ad(mtr_memo_contains_page(mtr, sp_header, MTR_MEMO_PAGE_SX_FIX));
 	ut_ad(page_offset(sp_header) == FSP_HEADER_OFFSET);
 	/* Read free limit and space size */
@@ -552,7 +552,7 @@ xdes_lst_get_descriptor(
 	fil_addr_t		lst_node,
 	mtr_t*			mtr)
 {
-	ut_ad(mtr_memo_contains(mtr, space, MTR_MEMO_SPACE_X_LOCK));
+	ut_ad(mtr_memo_contains(mtr, &space->latch, MTR_MEMO_X_LOCK));
 	ut_ad(page_size.equals_to(page_size_t(space->flags)));
 	return(fut_get_ptr(space->id, page_size, lst_node, RW_SX_LATCH, mtr)
 	       - XDES_FLST_NODE);
@@ -574,9 +574,13 @@ xdes_get_offset(
 	       * FSP_EXTENT_SIZE);
 }
 
-/** Initialize a file page whose prior contents should be ignored.
-@param[in,out]	block	buffer pool block */
-void fsp_apply_init_file_page(buf_block_t* block)
+/***********************************************************//**
+Inits a file page whose prior contents should be ignored. */
+static
+void
+fsp_init_file_page_low(
+/*===================*/
+	buf_block_t*	block)	/*!< in: pointer to a page */
 {
 	page_t*		page	= buf_block_get_frame(block);
 
@@ -598,36 +602,84 @@ void fsp_apply_init_file_page(buf_block_t* block)
 #ifdef UNIV_DEBUG
 /** Assert that the mini-transaction is compatible with
 updating an allocation bitmap page.
+@param[in]	id	tablespace identifier
 @param[in]	mtr	mini-transaction */
-void fil_space_t::modify_check(const mtr_t& mtr) const
+static
+void
+fsp_space_modify_check(
+	const fil_space_t*	space,
+	const mtr_t*		mtr)
 {
-	switch (mtr.get_log_mode()) {
+	switch (mtr->get_log_mode()) {
 	case MTR_LOG_SHORT_INSERTS:
 	case MTR_LOG_NONE:
 		/* These modes are only allowed within a non-bitmap page
 		when there is a higher-level redo log record written. */
-		ut_ad(purpose == FIL_TYPE_TABLESPACE
-		      || purpose == FIL_TYPE_TEMPORARY);
+		ut_ad(space->purpose == FIL_TYPE_TABLESPACE
+		      || space->purpose == FIL_TYPE_TEMPORARY);
 		break;
 	case MTR_LOG_NO_REDO:
-		ut_ad(purpose == FIL_TYPE_TEMPORARY
-		      || purpose == FIL_TYPE_IMPORT
-		      || my_atomic_loadlint(&redo_skipped_count)
-		      || is_being_truncated
-		      || srv_is_tablespace_truncated(id));
+		ut_ad(space->purpose == FIL_TYPE_TEMPORARY
+		      || space->purpose == FIL_TYPE_IMPORT
+		      || my_atomic_loadlint(&space->redo_skipped_count));
 		return;
 	case MTR_LOG_ALL:
-		/* We may only write redo log for a persistent
-		tablespace. */
-		ut_ad(purpose == FIL_TYPE_TABLESPACE);
-		ut_ad(mtr.is_named_space(id));
+		/* We may only write redo log for a persistent tablespace. */
+		ut_ad(space->purpose == FIL_TYPE_TABLESPACE);
+		ut_ad(mtr->is_named_space(space->id));
 		return;
 	}
 
-	ut_ad(!"invalid log mode");
+	ut_ad(0);
 }
+#endif /* UNIV_DEBUG */
+
+/** Initialize a file page.
+@param[in,out]	block	file page
+@param[in,out]	mtr	mini-transaction */
+static
+void
+fsp_init_file_page(buf_block_t* block, mtr_t* mtr)
+{
+	fsp_init_file_page_low(block);
+
+	mlog_write_initial_log_record(buf_block_get_frame(block),
+				      MLOG_INIT_FILE_PAGE2, mtr);
+}
+
+#ifdef UNIV_DEBUG
+static
+void
+fsp_init_file_page(const fil_space_t* space, buf_block_t* block, mtr_t* mtr)
+{
+	ut_d(fsp_space_modify_check(space, mtr));
+	ut_ad(space->id == block->page.id.space());
+	fsp_init_file_page(block, mtr);
+}
+#else /* UNIV_DEBUG */
+# define fsp_init_file_page(space, block, mtr) fsp_init_file_page(block, mtr)
 #endif
 
+/***********************************************************//**
+Parses a redo log record of a file page init.
+@return end of log record or NULL */
+byte*
+fsp_parse_init_file_page(
+/*=====================*/
+	byte*		ptr,	/*!< in: buffer */
+	byte*		end_ptr MY_ATTRIBUTE((unused)), /*!< in: buffer end */
+	buf_block_t*	block)	/*!< in: block or NULL */
+{
+	ut_ad(ptr != NULL);
+	ut_ad(end_ptr != NULL);
+
+	if (block) {
+		fsp_init_file_page_low(block);
+	}
+
+	return(ptr);
+}
+
 /**********************************************************************//**
 Writes the space id and flags to a tablespace header.  The flags contain
 row type, physical/compressed page size, and logical/uncompressed page
@@ -657,7 +709,7 @@ void fsp_header_init(fil_space_t* space, ulint size, mtr_t* mtr)
 	const page_id_t		page_id(space->id, 0);
 	const page_size_t	page_size(space->flags);
 
-	mtr_x_lock_space(space, mtr);
+	mtr_x_lock(&space->latch, mtr);
 	buf_block_t* block = buf_page_create(page_id, page_size, mtr);
 	buf_page_get(page_id, page_size, RW_SX_LATCH, mtr);
 	buf_block_dbg_add_level(block, SYNC_FSP_PAGE);
@@ -675,23 +727,23 @@ void fsp_header_init(fil_space_t* space, ulint size, mtr_t* mtr)
 
 	mlog_write_ulint(FSP_HEADER_OFFSET + FSP_SPACE_ID + block->frame,
 			 space->id, MLOG_4BYTES, mtr);
-	mlog_write_ulint(FSP_HEADER_OFFSET + FSP_NOT_USED + block->frame, 0,
-			 MLOG_4BYTES, mtr);
+	ut_ad(0 == mach_read_from_4(FSP_HEADER_OFFSET + FSP_NOT_USED
+				    + block->frame));
 	mlog_write_ulint(FSP_HEADER_OFFSET + FSP_SIZE + block->frame, size,
 			 MLOG_4BYTES, mtr);
-	mlog_write_ulint(FSP_HEADER_OFFSET + FSP_FREE_LIMIT + block->frame, 0,
-			 MLOG_4BYTES, mtr);
+	ut_ad(0 == mach_read_from_4(FSP_HEADER_OFFSET + FSP_FREE_LIMIT
+				    + block->frame));
 	mlog_write_ulint(FSP_HEADER_OFFSET + FSP_SPACE_FLAGS + block->frame,
 			 space->flags & ~FSP_FLAGS_MEM_MASK,
 			 MLOG_4BYTES, mtr);
-	mlog_write_ulint(FSP_HEADER_OFFSET + FSP_FRAG_N_USED + block->frame, 0,
-			 MLOG_4BYTES, mtr);
+	ut_ad(0 == mach_read_from_4(FSP_HEADER_OFFSET + FSP_FRAG_N_USED
+				    + block->frame));
 
-	flst_init(FSP_HEADER_OFFSET + FSP_FREE + block->frame, mtr);
-	flst_init(FSP_HEADER_OFFSET + FSP_FREE_FRAG + block->frame, mtr);
-	flst_init(FSP_HEADER_OFFSET + FSP_FULL_FRAG + block->frame, mtr);
-	flst_init(FSP_HEADER_OFFSET + FSP_SEG_INODES_FULL + block->frame, mtr);
-	flst_init(FSP_HEADER_OFFSET + FSP_SEG_INODES_FREE + block->frame, mtr);
+	flst_init(block, FSP_HEADER_OFFSET + FSP_FREE, mtr);
+	flst_init(block, FSP_HEADER_OFFSET + FSP_FREE_FRAG, mtr);
+	flst_init(block, FSP_HEADER_OFFSET + FSP_FULL_FRAG, mtr);
+	flst_init(block, FSP_HEADER_OFFSET + FSP_SEG_INODES_FULL, mtr);
+	flst_init(block, FSP_HEADER_OFFSET + FSP_SEG_INODES_FREE, mtr);
 
 	mlog_write_ull(FSP_HEADER_OFFSET + FSP_SEG_ID + block->frame, 1, mtr);
 
@@ -753,7 +805,7 @@ fsp_try_extend_data_file_with_pages(
 	ulint	size;
 
 	ut_a(!is_system_tablespace(space->id));
-	ut_d(space->modify_check(*mtr));
+	ut_d(fsp_space_modify_check(space, mtr));
 
 	size = mach_read_from_4(header + FSP_SIZE);
 	ut_ad(size == space->size_in_header);
@@ -785,7 +837,7 @@ fsp_try_extend_data_file(fil_space_t* space, fsp_header_t* header, mtr_t* mtr)
 		"ran out of space. Please add another file or use"
 		" 'autoextend' for the last file in setting";
 
-	ut_d(space->modify_check(*mtr));
+	ut_d(fsp_space_modify_check(space, mtr));
 
 	if (space->id == TRX_SYS_SPACE
 	    && !srv_sys_space.can_auto_extend_last_file()) {
@@ -857,7 +909,7 @@ fsp_try_extend_data_file(fil_space_t* space, fsp_header_t* header, mtr_t* mtr)
 	/* We ignore any fragments of a full megabyte when storing the size
 	to the space header */
 
-	space->size_in_header = ut_2pow_round(
+	space->size_in_header = ut_calc_align_down(
 		space->size, (1024 * 1024) / page_size.physical());
 
 	mlog_write_ulint(
@@ -948,7 +1000,7 @@ fsp_fill_free_list(
 	ulint	i;
 
 	ut_ad(page_offset(header) == FSP_HEADER_OFFSET);
-	ut_d(space->modify_check(*mtr));
+	ut_d(fsp_space_modify_check(space, mtr));
 
 	/* Check if we can fill free list from above the free list limit */
 	size = mach_read_from_4(header + FSP_SIZE);
@@ -1014,22 +1066,35 @@ fsp_fill_free_list(
 						 MLOG_2BYTES, mtr);
 			}
 
+			/* Initialize the ibuf bitmap page in a separate
+			mini-transaction because it is low in the latching
+			order, and we must be able to release its latch.
+			Note: Insert-Buffering is disabled for tables that
+			reside in the temp-tablespace. */
 			if (space->purpose != FIL_TYPE_TEMPORARY) {
+				mtr_t	ibuf_mtr;
+
+				mtr_start(&ibuf_mtr);
+				ibuf_mtr.set_named_space(space);
+
 				const page_id_t	page_id(
 					space->id,
 					i + FSP_IBUF_BITMAP_OFFSET);
 
 				block = buf_page_create(
-					page_id, page_size, mtr);
+					page_id, page_size, &ibuf_mtr);
 
 				buf_page_get(
-					page_id, page_size, RW_SX_LATCH, mtr);
+					page_id, page_size, RW_SX_LATCH,
+					&ibuf_mtr);
 
 				buf_block_dbg_add_level(block, SYNC_FSP_PAGE);
 
-				fsp_init_file_page(space, block, mtr);
+				fsp_init_file_page(space, block, &ibuf_mtr);
 
-				ibuf_bitmap_page_init(block, mtr);
+				ibuf_bitmap_page_init(block, &ibuf_mtr);
+
+				mtr_commit(&ibuf_mtr);
 			}
 		}
 
@@ -1202,8 +1267,11 @@ fsp_page_create(
 	} else {
 		rw_lock_sx_lock(&block->lock);
 	}
+	mutex_enter(&block->mutex);
 
 	buf_block_buf_fix_inc(block, __FILE__, __LINE__);
+
+	mutex_exit(&block->mutex);
 	mtr_memo_push(init_mtr, block, rw_latch == RW_X_LATCH
 		      ? MTR_MEMO_PAGE_X_FIX : MTR_MEMO_PAGE_SX_FIX);
 
@@ -1251,7 +1319,7 @@ fsp_alloc_free_page(
 	ulint		free;
 	const ulint	space_id = space->id;
 
-	ut_d(space->modify_check(*mtr));
+	ut_d(fsp_space_modify_check(space, mtr));
 	header = fsp_get_space_header(space, page_size, mtr);
 
 	/* Get the hinted descriptor */
@@ -1319,7 +1387,7 @@ fsp_alloc_free_page(
 		ut_a(!is_system_tablespace(space_id));
 		if (page_no >= FSP_EXTENT_SIZE) {
 			ib::error() << "Trying to extend a single-table"
-				" tablespace " << space->name << " , by single"
+				" tablespace " << space << " , by single"
 				" page(s) though the space size " << space_size
 				<< ". Page no " << page_no << ".";
 			return(NULL);
@@ -1340,7 +1408,7 @@ fsp_alloc_free_page(
 /** Frees a single page of a space.
 The page is marked as free and clean.
 @param[in,out]	space		tablespace
-@param[in]	offset		page number
+@param[in]	page_id		page id
 @param[in]	page_size	page size
 @param[in,out]	mtr		mini-transaction */
 static
@@ -1357,7 +1425,7 @@ fsp_free_page(
 	ulint		frag_n_used;
 
 	ut_ad(mtr);
-	ut_d(space->modify_check(*mtr));
+	ut_d(fsp_space_modify_check(space, mtr));
 
 	/* fprintf(stderr, "Freeing page %lu in space %lu\n", page, space); */
 
@@ -1406,8 +1474,7 @@ fsp_free_page(
 	const ulint	bit = offset % FSP_EXTENT_SIZE;
 
 	xdes_set_bit(descr, XDES_FREE_BIT, bit, TRUE, mtr);
-	/* xdes_init() should have set all XDES_CLEAN_BIT */
-	ut_ad(xdes_get_bit(descr, XDES_CLEAN_BIT, bit));
+	xdes_set_bit(descr, XDES_CLEAN_BIT, bit, TRUE, mtr);
 
 	frag_n_used = mtr_read_ulint(header + FSP_FRAG_N_USED, MLOG_4BYTES,
 				     mtr);
@@ -1451,7 +1518,7 @@ fsp_free_extent(
 	fsp_header_t*	header;
 	xdes_t*		descr;
 
-	ut_ad(mtr_memo_contains(mtr, space, MTR_MEMO_SPACE_X_LOCK));
+	ut_ad(mtr_memo_contains(mtr, &space->latch, MTR_MEMO_X_LOCK));
 
 	header = fsp_get_space_header(space, page_size, mtr);
 
@@ -1564,7 +1631,9 @@ fsp_alloc_seg_inode_page(
 	fsp_header_t*	space_header,
 	mtr_t*		mtr)
 {
+	fseg_inode_t*	inode;
 	buf_block_t*	block;
+	page_t*		page;
 
 	ut_ad(page_offset(space_header) == FSP_HEADER_OFFSET);
 	ut_ad(page_get_space_id(page_align(space_header)) == space->id);
@@ -1582,20 +1651,22 @@ fsp_alloc_seg_inode_page(
 	buf_block_dbg_add_level(block, SYNC_FSP_PAGE);
 	ut_ad(rw_lock_get_sx_lock_count(&block->lock) == 1);
 
-	mlog_write_ulint(block->frame + FIL_PAGE_TYPE, FIL_PAGE_INODE,
+	page = buf_block_get_frame(block);
+
+	mlog_write_ulint(page + FIL_PAGE_TYPE, FIL_PAGE_INODE,
 			 MLOG_2BYTES, mtr);
 
-#ifdef UNIV_DEBUG
-	const byte* inode = FSEG_ID + FSEG_ARR_OFFSET + block->frame;
-	for (ulint i = FSP_SEG_INODES_PER_PAGE(page_size); i--;
-	     inode += FSEG_INODE_SIZE) {
-		ut_ad(!mach_read_from_8(inode));
+	for (ulint i = 0; i < FSP_SEG_INODES_PER_PAGE(page_size); i++) {
+
+		inode = fsp_seg_inode_page_get_nth_inode(
+			page, i, page_size, mtr);
+
+		mlog_write_ull(inode + FSEG_ID, 0, mtr);
 	}
-#endif
 
 	flst_add_last(
 		space_header + FSP_SEG_INODES_FREE,
-		block->frame + FSEG_INODE_PAGE_NODE, mtr);
+		page + FSEG_INODE_PAGE_NODE, mtr);
 
 	return(true);
 }
@@ -1676,7 +1747,7 @@ fsp_free_seg_inode(
 	page_t*		page;
 	fsp_header_t*	space_header;
 
-	ut_d(space->modify_check(*mtr));
+	ut_d(fsp_space_modify_check(space, mtr));
 
 	page = page_align(inode);
 
@@ -1924,9 +1995,9 @@ fseg_create(
 	ut_ad(byte_offset + FSEG_HEADER_SIZE
 	      <= srv_page_size - FIL_PAGE_DATA_END);
 
-	mtr_x_lock_space(space, mtr);
+	mtr_x_lock(&space->latch, mtr);
 	const page_size_t	page_size(space->flags);
-	ut_d(space->modify_check(*mtr));
+	ut_d(fsp_space_modify_check(space, mtr));
 
 	if (page != 0) {
 		block = buf_page_get(page_id_t(space->id, page), page_size,
@@ -2105,7 +2176,7 @@ fseg_fill_free_list(
 
 	ut_ad(inode && mtr);
 	ut_ad(!((page_offset(inode) - FSEG_ARR_OFFSET) % FSEG_INODE_SIZE));
-	ut_d(space->modify_check(*mtr));
+	ut_d(fsp_space_modify_check(space, mtr));
 
 	reserved = fseg_n_reserved_pages_low(inode, &used, mtr);
 
@@ -2173,7 +2244,7 @@ fseg_alloc_free_extent(
 
 	ut_ad(!((page_offset(inode) - FSEG_ARR_OFFSET) % FSEG_INODE_SIZE));
 	ut_ad(mach_read_from_4(inode + FSEG_MAGIC_N) == FSEG_MAGIC_N_VALUE);
-	ut_d(space->modify_check(*mtr));
+	ut_d(fsp_space_modify_check(space, mtr));
 
 	if (flst_get_len(inode + FSEG_FREE) > 0) {
 		/* Segment free list is not empty, allocate from it */
@@ -2260,7 +2331,7 @@ fseg_alloc_free_page_low(
 	seg_id = mach_read_from_8(seg_inode + FSEG_ID);
 
 	ut_ad(seg_id);
-	ut_d(space->modify_check(*mtr));
+	ut_d(fsp_space_modify_check(space, mtr));
 	ut_ad(fil_page_get_type(page_align(seg_inode)) == FIL_PAGE_INODE);
 
 	reserved = fseg_n_reserved_pages_low(seg_inode, &used, mtr);
@@ -2629,7 +2700,7 @@ fsp_reserve_free_extents(
 	ut_ad(mtr);
 	*n_reserved = n_ext;
 
-	mtr_x_lock_space(space, mtr);
+	mtr_x_lock(&space->latch, mtr);
 	const page_size_t	page_size(space->flags);
 
 	space_header = fsp_get_space_header(space, page_size, mtr);
@@ -2774,6 +2845,8 @@ fseg_mark_page_used(
 @param[in,out]	space		tablespace
 @param[in]	offset		page number
 @param[in]	page_size	page size
+@param[in]	ahi		whether we may need to drop the adaptive
+hash index
 @param[in,out]	mtr		mini-transaction */
 static
 void
@@ -2782,6 +2855,9 @@ fseg_free_page_low(
 	fil_space_t*		space,
 	page_no_t		offset,
 	const page_size_t&	page_size,
+#ifdef BTR_CUR_HASH_ADAPT
+	bool			ahi,
+#endif /* BTR_CUR_HASH_ADAPT */
 	mtr_t*			mtr)
 {
 	xdes_t*	descr;
@@ -2789,13 +2865,23 @@ fseg_free_page_low(
 	ulint	state;
 	ib_id_t	descr_id;
 	ib_id_t	seg_id;
+	ulint	i;
 
 	ut_ad(seg_inode != NULL);
 	ut_ad(mtr != NULL);
 	ut_ad(mach_read_from_4(seg_inode + FSEG_MAGIC_N)
 	      == FSEG_MAGIC_N_VALUE);
 	ut_ad(!((page_offset(seg_inode) - FSEG_ARR_OFFSET) % FSEG_INODE_SIZE));
-	ut_d(space->modify_check(*mtr));
+	ut_d(fsp_space_modify_check(space, mtr));
+#ifdef BTR_CUR_HASH_ADAPT
+	/* Drop search system page hash index if the page is found in
+	the pool and is hashed */
+
+	if (ahi) {
+		btr_search_drop_page_hash_when_freed(
+			page_id_t(space->id, offset));
+	}
+#endif /* BTR_CUR_HASH_ADAPT */
 
 	descr = xdes_get_descriptor(space, offset, page_size, mtr);
 
@@ -2815,7 +2901,7 @@ fseg_free_page_low(
 	if (state != XDES_FSEG) {
 		/* The page is in the fragment pages of the segment */
 
-		for (ulint i = 0;; i++) {
+		for (i = 0;; i++) {
 			if (fseg_get_nth_frag_page_no(seg_inode, i, mtr)
 			    == offset) {
 
@@ -2870,8 +2956,7 @@ fseg_free_page_low(
 	const ulint	bit = offset % FSP_EXTENT_SIZE;
 
 	xdes_set_bit(descr, XDES_FREE_BIT, bit, TRUE, mtr);
-	/* xdes_init() should have set all XDES_CLEAN_BIT */
-	ut_ad(xdes_get_bit(descr, XDES_CLEAN_BIT, bit));
+	xdes_set_bit(descr, XDES_CLEAN_BIT, bit, TRUE, mtr);
 
 	if (xdes_is_free(descr, mtr)) {
 		/* The extent has become free: free it to space */
@@ -2881,34 +2966,40 @@ fseg_free_page_low(
 	}
 }
 
-/** Free a page in a file segment.
-@param[in,out]	seg_header	file segment header
-@param[in,out]	space		tablespace
-@param[in]	offset		page number
-@param[in,out]	mtr		mini-transaction */
+#ifndef BTR_CUR_HASH_ADAPT
+# define fseg_free_page_low(inode, space, offset, page_size, ahi, mtr)	\
+	fseg_free_page_low(inode, space, offset, page_size, mtr)
+#endif /* !BTR_CUR_HASH_ADAPT */
+
+/**********************************************************************//**
+Frees a single page of a segment. */
 void
-fseg_free_page(
-	fseg_header_t*	seg_header,
-	fil_space_t*	space,
-	ulint		offset,
-	mtr_t*		mtr)
+fseg_free_page_func(
+	fseg_header_t*	seg_header, /*!< in: segment header */
+	ulint		space_id,/*!< in: space id */
+	ulint		page,	/*!< in: page offset */
+#ifdef BTR_CUR_HASH_ADAPT
+	bool		ahi,	/*!< in: whether we may need to drop
+				the adaptive hash index */
+#endif /* BTR_CUR_HASH_ADAPT */
+	mtr_t*		mtr)	/*!< in/out: mini-transaction */
 {
 	DBUG_ENTER("fseg_free_page");
 	fseg_inode_t*		seg_inode;
 	buf_block_t*		iblock;
-	mtr_x_lock_space(space, mtr);
+	fil_space_t*		space = mtr_x_lock_space(space_id, mtr);
 	const page_size_t	page_size(space->flags);
 
-	DBUG_LOG("fseg_free_page", "space_id: " << space->id
-		 << ", page_no: " << offset);
+	DBUG_LOG("fseg_free_page", "space_id: " << space_id
+		 << ", page_no: " << page);
 
-	seg_inode = fseg_inode_get(seg_header, space->id, page_size, mtr,
+	seg_inode = fseg_inode_get(seg_header, space_id, page_size, mtr,
 				   &iblock);
 	fil_block_check_type(*iblock, FIL_PAGE_INODE, mtr);
 
-	fseg_free_page_low(seg_inode, space, offset, page_size, mtr);
+	fseg_free_page_low(seg_inode, space, page, page_size, ahi, mtr);
 
-	ut_d(buf_page_set_file_page_was_freed(page_id_t(space->id, offset)));
+	ut_d(buf_page_set_file_page_was_freed(page_id_t(space_id, page)));
 
 	DBUG_VOID_RETURN;
 }
@@ -2926,7 +3017,7 @@ fseg_page_is_free(fil_space_t* space, unsigned page)
 	page_no_t	dpage = xdes_calc_descriptor_page(page_size, page);
 
 	mtr.start();
-	mtr_s_lock_space(space, &mtr);
+	mtr_s_lock(&space->latch, &mtr);
 
 	if (page >= space->free_limit || page >= space->size_in_header) {
 		is_free = true;
@@ -2947,6 +3038,8 @@ fseg_page_is_free(fil_space_t* space, unsigned page)
 @param[in,out]	space		tablespace
 @param[in]	page_size	page size
 @param[in]	page		page number in the extent
+@param[in]	ahi		whether we may need to drop
+				the adaptive hash index
 @param[in,out]	mtr		mini-transaction */
 MY_ATTRIBUTE((nonnull))
 static
@@ -2956,11 +3049,16 @@ fseg_free_extent(
 	fil_space_t*		space,
 	const page_size_t&	page_size,
 	ulint			page,
+#ifdef BTR_CUR_HASH_ADAPT
+	bool			ahi,
+#endif /* BTR_CUR_HASH_ADAPT */
 	mtr_t*			mtr)
 {
+	ulint	first_page_in_extent;
 	xdes_t*	descr;
 	ulint	not_full_n_used;
 	ulint	descr_n_used;
+	ulint	i;
 
 	ut_ad(mtr != NULL);
 
@@ -2970,8 +3068,26 @@ fseg_free_extent(
 	ut_a(!memcmp(descr + XDES_ID, seg_inode + FSEG_ID, 8));
 	ut_ad(mach_read_from_4(seg_inode + FSEG_MAGIC_N)
 	      == FSEG_MAGIC_N_VALUE);
-	ut_d(space->modify_check(*mtr));
-	ut_d(ulint first_page_in_extent = page - (page % FSP_EXTENT_SIZE));
+	ut_d(fsp_space_modify_check(space, mtr));
+
+	first_page_in_extent = page - (page % FSP_EXTENT_SIZE);
+
+#ifdef BTR_CUR_HASH_ADAPT
+	if (ahi) {
+		for (i = 0; i < FSP_EXTENT_SIZE; i++) {
+			if (!xdes_mtr_get_bit(descr, XDES_FREE_BIT, i, mtr)) {
+
+				/* Drop search system page hash index
+				if the page is found in the pool and
+				is hashed */
+
+				btr_search_drop_page_hash_when_freed(
+					page_id_t(space->id,
+						  first_page_in_extent + i));
+			}
+		}
+	}
+#endif /* BTR_CUR_HASH_ADAPT */
 
 	if (xdes_is_full(descr, mtr)) {
 		flst_remove(seg_inode + FSEG_FULL,
@@ -2996,7 +3112,7 @@ fseg_free_extent(
 	fsp_free_extent(space, page, page_size, mtr);
 
 #ifdef UNIV_DEBUG
-	for (ulint i = 0; i < FSP_EXTENT_SIZE; i++) {
+	for (i = 0; i < FSP_EXTENT_SIZE; i++) {
 
 		buf_page_set_file_page_was_freed(
 			page_id_t(space->id, first_page_in_extent + i));
@@ -3004,18 +3120,27 @@ fseg_free_extent(
 #endif /* UNIV_DEBUG */
 }
 
+#ifndef BTR_CUR_HASH_ADAPT
+# define fseg_free_extent(inode, space, page_size, page, ahi, mtr)	\
+	fseg_free_extent(inode, space, page_size, page, mtr)
+#endif /* !BTR_CUR_HASH_ADAPT */
+
 /**********************************************************************//**
 Frees part of a segment. This function can be used to free a segment by
 repeatedly calling this function in different mini-transactions. Doing
 the freeing in a single mini-transaction might result in too big a
 mini-transaction.
-@return whether the freeing was completed */
-bool
-fseg_free_step(
+@return TRUE if freeing completed */
+ibool
+fseg_free_step_func(
 	fseg_header_t*	header,	/*!< in, own: segment header; NOTE: if the header
 				resides on the first page of the frag list
 				of the segment, this pointer becomes obsolete
 				after the last freeing step */
+#ifdef BTR_CUR_HASH_ADAPT
+	bool		ahi,	/*!< in: whether we may need to drop
+				the adaptive hash index */
+#endif /* BTR_CUR_HASH_ADAPT */
 	mtr_t*		mtr)	/*!< in/out: mini-transaction */
 {
 	ulint		n;
@@ -3047,7 +3172,7 @@ fseg_free_step(
 	if (inode == NULL) {
 		ib::info() << "Double free of inode from "
 			<< page_id_t(space_id, header_page);
-		DBUG_RETURN(true);
+		DBUG_RETURN(TRUE);
 	}
 
 	fil_block_check_type(*iblock, FIL_PAGE_INODE, mtr);
@@ -3057,9 +3182,9 @@ fseg_free_step(
 		/* Free the extent held by the segment */
 		page = xdes_get_offset(descr);
 
-		fseg_free_extent(inode, space, page_size, page, mtr);
+		fseg_free_extent(inode, space, page_size, page, ahi, mtr);
 
-		DBUG_RETURN(false);
+		DBUG_RETURN(FALSE);
 	}
 
 	/* Free a frag page */
@@ -3069,13 +3194,13 @@ fseg_free_step(
 		/* Freeing completed: free the segment inode */
 		fsp_free_seg_inode(space, page_size, inode, mtr);
 
-		DBUG_RETURN(true);
+		DBUG_RETURN(TRUE);
 	}
 
 	fseg_free_page_low(
 		inode, space,
 		fseg_get_nth_frag_page_no(inode, n, mtr),
-		page_size, mtr);
+		page_size, ahi, mtr);
 
 	n = fseg_find_last_used_frag_page_slot(inode, mtr);
 
@@ -3083,20 +3208,24 @@ fseg_free_step(
 		/* Freeing completed: free the segment inode */
 		fsp_free_seg_inode(space, page_size, inode, mtr);
 
-		DBUG_RETURN(true);
+		DBUG_RETURN(TRUE);
 	}
 
-	DBUG_RETURN(false);
+	DBUG_RETURN(FALSE);
 }
 
 /**********************************************************************//**
 Frees part of a segment. Differs from fseg_free_step because this function
 leaves the header page unfreed.
-@return whether the freeing was completed, except for the header page */
-bool
-fseg_free_step_not_header(
+@return TRUE if freeing completed, except the header page */
+ibool
+fseg_free_step_not_header_func(
 	fseg_header_t*	header,	/*!< in: segment header which must reside on
 				the first fragment page of the segment */
+#ifdef BTR_CUR_HASH_ADAPT
+	bool		ahi,	/*!< in: whether we may need to drop
+				the adaptive hash index */
+#endif /* BTR_CUR_HASH_ADAPT */
 	mtr_t*		mtr)	/*!< in/out: mini-transaction */
 {
 	ulint		n;
@@ -3122,27 +3251,29 @@ fseg_free_step_not_header(
 		/* Free the extent held by the segment */
 		page = xdes_get_offset(descr);
 
-		fseg_free_extent(inode, space, page_size, page, mtr);
+		fseg_free_extent(inode, space, page_size, page, ahi, mtr);
 
-		return false;
+		return(FALSE);
 	}
 
 	/* Free a frag page */
 
 	n = fseg_find_last_used_frag_page_slot(inode, mtr);
 
-	ut_a(n != ULINT_UNDEFINED);
+	if (n == ULINT_UNDEFINED) {
+		ut_error;
+	}
 
 	page_no = fseg_get_nth_frag_page_no(inode, n, mtr);
 
 	if (page_no == page_get_page_no(page_align(header))) {
 
-		return true;
+		return(TRUE);
 	}
 
-	fseg_free_page_low(inode, space, page_no, page_size, mtr);
+	fseg_free_page_low(inode, space, page_no, page_size, ahi, mtr);
 
-	return false;
+	return(FALSE);
 }
 
 /** Returns the first extent descriptor for a segment.
